//
// CUDA benchmark for running a STREAM-like benchmark on GPUs to determine bandwidth vs. thread configuration
//
// Author: Karl Rupp,  me@karlrupp.net
// License: MIT/X11 license, see file LICENSE.txt
//

#include <iostream>
#include <string>

#ifndef VIENNACL_WITH_CUDA
  #define VIENNACL_WITH_CUDA
#endif

#include <cuda_runtime.h>

// ViennaCL headers
#include "viennacl/vector.hpp"
#include "viennacl/tools/timer.hpp"


inline void cuda_last_error_check()
{
  cudaError_t error_code = cudaGetLastError();

  if (cudaSuccess != error_code)
  {
    std::stringstream ss;
    ss << "CUDA Runtime API error " << error_code << ": " << cudaGetErrorString( error_code ) << std::endl;
    throw std::runtime_error(ss.str());
  }
}


// Kernel for the benchmark
template<typename NumericT>
__global__ void elementwise_add(const NumericT * x,
                                const NumericT * y,
                                      NumericT * z,
                                unsigned int size)
{
  for (unsigned int i = blockDim.x * blockIdx.x + threadIdx.x;
                    i < size;
                    i += gridDim.x * blockDim.x)
    z[i] = x[i] + y[i];
}


int main(int argc, char **argv)
{
  typedef float       NumericT;

  unsigned int threadblocksize = 128;
  if (argc > 1)
    threadblocksize = atoi(argv[1]);

  cudaDeviceProp prop;
  cudaGetDeviceProperties(&prop, 0);
  std::cout << "# Using device: " << prop.name << std::endl;

  // Set up work vectors
  std::size_t N = 5000000;
  viennacl::vector<NumericT>  x(N);
  viennacl::vector<NumericT>  y(N);
  viennacl::vector<NumericT>  z(N);


  // Warmup calculation:
  elementwise_add<<<128, threadblocksize>>>(viennacl::cuda_arg(x),
                                            viennacl::cuda_arg(y),
                                            viennacl::cuda_arg(z),
                                            static_cast<unsigned int>(x.size()));
  cuda_last_error_check();

  // Benchmark runs
  viennacl::tools::timer timer;
  std::cout << "# Starting benchmark runs with " << threadblocksize << " threads per thread block." << std::endl;
  if (argc == 1)
    std::cout << "# Note: A different thread block size can be provided as command line argument" << std::endl;
  std::cout << "# num_threadblocks     time       GB/sec" << std::endl;
  for (std::size_t num_threadblocks = 1; num_threadblocks <= 64; num_threadblocks *= 2)
  {
    viennacl::backend::finish();
    timer.start();

    // repeat calculation several times, then average
    for (std::size_t num_runs = 0; num_runs < 20; ++num_runs)
    {
      elementwise_add<<<num_threadblocks, threadblocksize>>>(viennacl::cuda_arg(x),
                                                             viennacl::cuda_arg(y),
                                                             viennacl::cuda_arg(z),
                                                             static_cast<unsigned int>(x.size()));
      cuda_last_error_check();
    }
    viennacl::backend::finish();
    double exec_time = timer.get();

    std::cout << "        " << num_threadblocks << "        " << exec_time << "        " <<  20.0 * 3.0 * sizeof(NumericT) * x.size() / exec_time * 1e-9 << std::endl;
  }

  return EXIT_SUCCESS;
}

