//
// OpenCL benchmark for running a STREAM-like benchmark on GPUs to determine bandwidth vs. workgroup configuration
//
// Author: Karl Rupp,  me@karlrupp.net
// License: MIT/X11 license, see file LICENSE.txt
//

#include <iostream>
#include <string>

#ifndef VIENNACL_WITH_OPENCL
  #define VIENNACL_WITH_OPENCL
#endif

// ViennaCL headers
#include "viennacl/ocl/backend.hpp"
#include "viennacl/vector.hpp"
#include "viennacl/tools/timer.hpp"


// Kernel for the benchmark
static const char * benchmark_program =
"__kernel void elementwise_add(\n"
"          __global const float * x,\n"
"          __global const float * y, \n"
"          __global float * z,\n"
"          unsigned int size) \n"
"{ \n"
"  for (unsigned int i = get_global_id(0); i < size; i += get_global_size(0))\n"
"    z[i] = x[i] + y[i];\n"
"};\n";

int main(int argc, char **argv)
{
  typedef float       NumericT;

  unsigned int workgroupsize = 128;
  if (argc > 1)
    workgroupsize = atoi(argv[1]);

  // Set up work vectors
  std::size_t N = 5000000;
  viennacl::vector<NumericT>  x(N);
  viennacl::vector<NumericT>  y(N);
  viennacl::vector<NumericT>  z(N);

  // Compile and extract kernel:
  viennacl::ocl::program & my_prog = viennacl::ocl::current_context().add_program(benchmark_program, "benchmark_program");
  viennacl::ocl::kernel & benchmark_kernel = my_prog.get_kernel("elementwise_add");

  // Warmup calculation:
  benchmark_kernel.local_work_size(0,       workgroupsize);
  benchmark_kernel.global_work_size(0, 128 * workgroupsize);
  viennacl::ocl::enqueue(benchmark_kernel(x, y, z, static_cast<cl_uint>(x.size())));

  // Benchmark runs
  viennacl::tools::timer timer;
  std::cout << "# Using device: " << viennacl::ocl::current_device().name() << std::endl;
  std::cout << "# Starting benchmark runs with " << workgroupsize << " threads per workgroup." << std::endl;
  if (argc == 1)
    std::cout << "# Note: A different workgroup size can be provided as command line argument" << std::endl;
  std::cout << "# num_workgroups     time       GB/sec" << std::endl;
  for (std::size_t num_workgroups = 1; num_workgroups <= 64; num_workgroups *= 2)
  {
    viennacl::backend::finish();
    timer.start();
    benchmark_kernel.local_work_size(0, workgroupsize);
    benchmark_kernel.global_work_size(0, num_workgroups * workgroupsize);

    // repeat calculation several times, then average
    for (std::size_t num_runs = 0; num_runs < 20; ++num_runs)
      viennacl::ocl::enqueue(benchmark_kernel(x, y, z, static_cast<cl_uint>(x.size())));
    viennacl::backend::finish();
    double exec_time = timer.get();

    std::cout << "        " << num_workgroups << "        " << exec_time << "        " <<  20.0 * 3.0 * sizeof(NumericT) * x.size() / exec_time * 1e-9 << std::endl;
  }

  return EXIT_SUCCESS;
}

